#ifndef COMMON_H
#define COMMON_H

#include "../../config.h"

extern time_t startTime;
extern double lat, lon;

/* Slovo je dvojice (retezec pismen, slovni tridou). */
class Word {
public:
  string text;       // Retezec pismen
  unsigned category; // Cislo slovni tridy, do ktereho slovo patri.
  Word (const string t, unsigned c) : text(t), category(c) {}
};

/*
  Trida obsahujici castecna reseni a slouzici k porovnavani jejich kvality.
  Jedna instance je ekvivalentni jednomu castecnemu reseni a ve statickych
  promennych tridy se uchovava seznam nejlepsich dosud nalezenych kompletnich
  reseni.
*/
class Solution {
public:
  /* Nejlepsi nalezena reseni uchovavame v hlade. */
  static priority_queue<Solution> all;
  /* Funkce, ktera vypise nejlepsi nalezena reseni v prehlednem formatu. */
  static void print_all ();

  /* Oznaci aktualni reseni za hotove a vyhodnoti jeho celkovou cenu */
  void complete ();

  vector <const Word *> words; // jednotliva slova castecneho reseni
  int cost; // cena aktualniho castecneho reseni

  // Lepsi reseni je to s mensi cenou
  bool operator<  (const Solution &) const;
  bool operator<= (const Solution &) const;
  bool operator>  (const Solution &) const;
  bool operator>= (const Solution &) const;
  bool operator== (const Solution &) const;

  void push_back (const Word *); // pridani slova na konec
  void pop_back ();              // odebrani slova z konce
  void clear ();                 // vyprazdni reseni
  size_t size () const;          // vraci pocet slov reseni

  Solution ();

  friend ostream & operator<< (ostream &, const Solution &);
};

/*
  Trida spravujici kontextova pravidla jako podminene ceny.
  Metoda get(rule_code) vrati cenu slova z posledni slovni tridy,
  kterou je potreba zaplatit za rozsireni (n-1)-gramu.
  Kazde pravidlo je reprezentovano jednim nezapornym celym cislem.

  Slovni tridy: 0, 1, 2, ..., Q - 1
  Pravidlo = T1, T2, T3, ... TL, kde Ti je cislo i-te slovni tridy v pravidlu
  Kod pravidla = T1 * Q ^ (L - 1) + T2 * Q ^ (L - 2) + ... + T(L-1) * Q + TL
*/
class Rules {
private:
  static map<unsigned, unsigned> rules;
  static map<string, unsigned> categories_map;
public:
  static size_t categories_count();              // Vrati pocet slovnich trid
  static unsigned add_category(string category); // Vytvori novou slovni tridu
  static unsigned get_category(string category); // Vrati cislo slovni tridy
  static void load();  // Nacte pravidla ze souboru
  static unsigned get(unsigned rule_code); // Nalezne cenu zadaneho pravidla
};

/*
  Jedna instance nasledujici tridy uchovava rozsireni nejakeho castecneho reseni.
  Jednotlive rozsireni je mozne porovnavat k dosazeni vhodneho poradi prochazeni.
*/
class Option {
public:
  const Word * word;
  int cost;
  bool operator< (const Option &) const;
  Option (const Word * _word) : word(_word) {}
  Option (const Word * _word, const Solution &);
};

/*
   Trida slouzici ke sprave prevodni tabulky - na zacatku je prazdna a meni
   se podle toho, kolik textu uz bylo uspesne prelozeno.
*/
class ConversionTable {
private:
  char conversion_table[P];
  bool used_letter[P];
public:
  void clear ();
  ConversionTable ();
  ConversionTable & operator= (const ConversionTable &);
  /* Od pozice start v retezci string pridava slovo word a vraci, zda uspela */
  int match (const string & cipher, int start, const Option &, const Solution &);
  char get (char) const;
  bool set (char, char);

  friend ostream & operator<< (ostream &, const ConversionTable &);
};

/* Trida slouzici se sprave slovniku. */
class Wordlist {
private:
  // trida je singleton, a proto ma privatni konstruktor
  static Wordlist* _instance;
  Wordlist ();

  static int trie[MAX_TRIE_SIZE][2*P];
  static int trie_size;
  static vector<const Word *> words[MAX_TRIE_SIZE];

  void add_to_trie (const Word *) ;
  void generate (Word * word, size_t pos, char n) ;

  void load (string filename, unsigned category);
  void _get_words (const string & cipher, size_t position,
                   const ConversionTable &, const Solution &,
                   vector<Option> & result) const;

public:
  static void get_words (const string & cipher, size_t position,
                         const ConversionTable &, const Solution &,
                         vector<Option> & result);
  void load_dir ();
  void load_places ();
  static Wordlist * Instance ();
  static Word start_word, end_word;
};

/* Nacte ze standartniho vstupu jeden radek, prevede ho na mala pismena a vrati. */
string get_input ();

/* Funkce pro vypis aktualniho stavu prohledavani. */
void progress (const string & input, const Solution &,
               const vector<Option> &, const ConversionTable &);

#endif
