#include "common.h"

// Preklada retezcove nazvy slovnich trid na cisla
map<string, unsigned> Rules::categories_map;
map<unsigned, unsigned> Rules::rules; // Kazdemu pravidlu je prirazena cena

size_t Rules::categories_count() { return Rules::categories_map.size(); }

unsigned Rules::add_category(string category) {
  unsigned id = categories_count();
  categories_map[category] = id;
  return id;
}

unsigned Rules::get_category(string category) {
  return categories_map[category];
}

// Nacte pravidla ze souboru
void Rules::load() {
  ifstream rules_file;
  rules_file.open(rules_config_filename.c_str());
  unsigned rule_length;
  while ( rules_file >> rule_length) {
    int cost;
    rules_file >> cost;
    vector<int> categories_list;
    for (size_t i = 0; i < rule_length; ++i) {
      string category;
      rules_file >> category;
      categories_list.push_back(categories_map[category]);
    }

    if (rule_length > N_MAX) continue;

    unsigned rule_code = 0;
    for (int i = rule_length - 1; i >= 0; --i) {
      rule_code = rule_code * (categories_count() + 1) + categories_list[i] + 1;
    }

    rules[rule_code] = cost + COST_DIFF;
  }

  rules_file.close();
}

unsigned Rules::get(unsigned rule_code) {
  if (rules.find(rule_code) != rules.end()) return rules[rule_code];
  return 0;
}

// Heuristika na poradi rozsirujicich slov
bool Option::operator< (const Option & op) const {
  switch (ORDER) {
    case ORDER_C: return cost < op.cost;
    case ORDER_L: return word->text.size() > op.word->text.size();
    case ORDER_CL:
      return (cost * (1 + op.word->text.size()) <
              op.cost * (1 + word->text.size()));
  }
}

// Konstruktor rozsirujiciho slova
Option::Option (const Word * w, const Solution & solution) : word(w) {
  unsigned categories_count = Rules::categories_count();
  unsigned rule_code = w->category + 1;
  size_t position = solution.size();
  while (true) {
    unsigned new_category;
    if (position <= 0) new_category = Wordlist::start_word.category;
    else new_category = solution.words[--position]->category;
    unsigned new_rule_code = rule_code * (categories_count + 1) + new_category + 1;
    if (!Rules::get(new_rule_code)) break;
    rule_code = new_rule_code;
  }
  cost = Rules::get(rule_code);
}

// Kazdy vrchol trie je reprezentovan polem "ukazatelu" delky 2P
vector<const Word *> Wordlist::words[MAX_TRIE_SIZE];
int Wordlist::trie[MAX_TRIE_SIZE][2*P], Wordlist::trie_size = 1;

Word Wordlist::start_word = Word("", 0), Wordlist::end_word = Word("", 1);

Wordlist::Wordlist () {}

Wordlist * Wordlist::_instance = NULL;

Wordlist * Wordlist::Instance () {
  if (!_instance) _instance = new Wordlist;
  return _instance;
}

/* Prida zadane slovo do trie */
void Wordlist::add_to_trie (const Word * word) {
  const string & text = word->text;
  int t = 0;
  for (size_t i = 0; i < text.size(); i++) {
    int n = text[i] - 'A';
    if (text[i] >= 'a' && text[i] <= 'z') n = P + text[i] - 'a';
    if (trie[t][n]) t = trie[t][n];
    else t = trie[t][n] = trie_size++;
  }
  words[t].push_back(word);
}

// kazda metoda, ktera tuto promennou pouziva, tak ji musi zanechat vynulovanou
static char translate[2*P];

/*
   Pro zadane slovo vygeneruje vsechny mozne varianty.
   Kazde z jeho pismen muze byt bud zname nebo nezname.
   Neznama pismena jsou nahrazena pismeny 'a', 'b', ...
   napr. ze slova XYX jsou vygenerovany XYX, aYa, XaX, aba
*/
void Wordlist::generate (Word * word, size_t pos, char n) {
  if (pos == word->text.size()) { add_to_trie (word); return; }
  int c = word->text[pos]-'A';
  if (translate[c]) {
    word->text[pos] = translate[c];
    generate (word, pos + 1, n);
    word->text[pos] = c + 'A';
    return;
  }
  word->text[pos] = translate[c] = n;
  generate (word, pos + 1, n + 1);
  word->text[pos] = translate[c] = c + 'A';
  // moc dlouha slova maji moc podmnozin a vsechny se bohuzel nevejdou do pameti
  // uchovame tedy u nich pouze variantu, ze nezname jeste zadny preklad
  if (word->text.size() <= MAX_LENGTH_TO_GENERATE_VARIANTS) generate (word, pos + 1, n);
  translate[c] = 0;
}

/*
  Tato metoda nacte vsechny kategorie slovniku a pravidla
  pro jejich pravdepodobnosti.
*/
void Wordlist::load_dir () {
  ifstream categories;
  categories.open(categories_config_filename.c_str());

  Rules::add_category("^");
  Rules::add_category("$");

  string category;
  while ( categories >> category) {
    if (!lat || (category != "ulice" && category != "obec"))
      load(wordlist_dir + category, Rules::categories_count());
    Rules::add_category(category);
  }

  categories.close();

  Wordlist::load_places();
}

void Wordlist::load_places() {

  if (!lat) return;

  ifstream file;
  file.open(geography_filename.c_str());

  int places_nearby = 0;

  double _lat, _lon;
  string type, name;
  while (file >> _lat) {
    file >> _lon;
    file >> type;
    file >> name;

    // Vzdalenost dvou bodu na zemi zadane zemepisnymi souradnicemi
    double lat1 = lat / 180 * M_PI, lat2 = _lat / 180 * M_PI;
    double dlon = (lon - _lon) / 180 * M_PI;
    double d = acos(sin(lat1) * sin(lat2) + cos(lat1) * cos(lat2) * cos(dlon)) * EARTH_RADIUS;
    if (d > MAX_DISTANCE) continue;

    int category;
    if (type == "village") category = Rules::get_category("obec");
    else category = Rules::get_category("ulice");

    places_nearby++;
    Word * word = new Word (name, category);
    generate (word, 0, 'a');
  }

  if (PROGRESS) cerr << "Places nearby: " << places_nearby << endl;
}

/* Nacte nova slova ze zadaneho souboru a vlozi je do trie. */
void Wordlist::load (string filename, unsigned category) {
  ifstream file;
  file.open(filename.c_str());

  string text;
  while ( file >> text ) {
    string text_only_letters;
    for (size_t i = 0; i < text.size(); i++) {
      if (text[i] >= 'A' && text[i] <= 'Z') {
        text_only_letters.push_back(text[i]);
      }
    }

    Word * word = new Word (text_only_letters, category);
    generate (word, 0, 'a');
  }

  file.close();
}

/*
  K zadane sifre (string cipher) vraci seznam slov (vector result), kterymi muze
  rozsifrovany text pokracovat od zadane pozice pozice (size_t position) pokud
  vezmeme v uvahu castecne vyplnenou prekladovou tabulku (ConversionTable conversion_table)

  Nezarucuje, ze vsemi vracenymi slovy lze skutecne pokracovat, ale jiste zadne slovo
  nevynecha. Snazi se tedy alespon minimalizovat pocet chybnych vyskytu.
*/
void Wordlist::get_words (const string & cipher,
                          size_t position,
                          const ConversionTable & conversion_table,
                          const Solution & solution,
                          vector<Option> & result) {
  Wordlist::Instance()->_get_words(cipher, position,
                                   conversion_table, solution, result);
}
void Wordlist::_get_words (const string & cipher,
                           size_t position,
                           const ConversionTable & conversion_table,
                           const Solution & solution,
                           vector<Option> & result) const {
  result.clear();

  // slova delsi nez MAX_LENGTH_TO_GENERATE_VARIANTS filtruje pouze podle jejich normalizovane formy
  // napr. KOLOBEZKA -> abcbdefag
  int node = 0;
  char n = 'a';
  for (size_t i = position; i < cipher.size(); i++) {
    int c = cipher[i] - 'a';

    if (!translate[c]) translate[c] = n++;
    node = trie[node][P + translate[c] - 'a'];

    if (!node) break;
    if (i - position + 1 > MAX_LENGTH_TO_GENERATE_VARIANTS) {
      for (size_t j = 0; j < words[node].size(); j++) {
        result.push_back(Option(words[node][j], solution));
      }
    }
  }
  for (int i = 0; i < 2*P; i++) translate[i] = 0;

  /*
    Pro slova dostatecne kratka (<= MAX_LENGTH_TO_GENERATE_VARIANTS) bere v uvahu
    i dosavadni prekladovou tabulku: vice vyplnenych pismen -> mene vysledku
  */
  node = 0;
  n = 'a';
  for (size_t i = position; i < cipher.size(); i++) {
    int c = cipher[i] - 'a';
    if (conversion_table.get(cipher[i])) {
      node = trie[node][conversion_table.get(cipher[i])-'A'];
    }
    else {
      if (!translate[c]) translate[c] = n++;
      node = trie[node][P + translate[c] - 'a'];
    }

    if (!node || i - position + 1 > MAX_LENGTH_TO_GENERATE_VARIANTS) break;
    for (size_t j = 0; j < words[node].size(); j++) {
      result.push_back(Option(words[node][j], solution));
    }
  }
  for (int i = 0; i < 2*P; i++) translate[i] = 0;

  // Setrizeni rozsirujicich slov podle zvolene heuristiky
  sort (result.begin(), result.end());
}
